<?php
/**
 * Piwik - Open source web analytics
 *
 * @author Jan Bogutzki
 * @link http://jan.bogutzki.de
 * @license http://www.gnu.org/licenses/gpl-3.0.html Gpl v3 or later
 * @version 1.2.1.1
 *
 * @package Piwik_Forecast
 */

/**
 *
 * @package Piwik_Forecast
 */
class Piwik_Forecast_API
{
	static private $instance = null;
	private $timefactor_nb_visits = 0;
	private $timefactor_nb_actions = 0;
	private $timefactor_bounce_count = 0;
	private $timefactor_nb_visits_returning = 0;
	private $today = 0;

	static public function getInstance()
	{
		if (self::$instance == null)
		{
			$c = __CLASS__;
			self::$instance = new $c();
		}
		return self::$instance;
	}

	public function getForecast( $idSite, $period, $date )
	{
		Piwik::checkUserHasViewAccess( $idSite );

		$dataTable = new Piwik_DataTable();

		$this->initForecast($idSite);

		// Unique Visits
		$arr = array('Forecast_Column'=>$this->getForecastValues($idSite, 'nb_visits'), 'label'=>Piwik_Translate('Forecast_Visitors'));
		$dataTable->addRowsFromSimpleArray($arr);

		// Pageviews
		$arr = array('Forecast_Column'=>$this->getForecastValues($idSite, 'nb_actions'), 'label'=>Piwik_Translate('Forecast_Actions'));
		$dataTable->addRowsFromSimpleArray($arr);

		// Bounces
		$arr = array('Forecast_Column'=>$this->getForecastValues($idSite, 'bounce_count'), 'label'=>Piwik_Translate('Forecast_Bounce'));
		$dataTable->addRowsFromSimpleArray($arr);

		// Returning
		$arr = array('Forecast_Column'=>$this->getForecastValues($idSite, 'nb_visits_returning'), 'label'=>Piwik_Translate('Forecast_ReturnVisits'));
		$dataTable->addRowsFromSimpleArray($arr);

		return $dataTable;
	}

	private function initForecast($idSite) {
		$this->today	= strtotime('today');
		$lastweek		= strtotime('today last week');

		// Fetch todays hourly visits
		$aToday = $this->getBlob($idSite, 'day', date('Y-m-d', $this->today), 'VisitTime_serverTime');

		// Fetch the last weeks hourly visits of that workday
		$aLastweek = $this->getBlob($idSite, 'day', date('Y-m-d', $lastweek), 'VisitTime_serverTime');

		// calculate the gradient
		$lastFullHour = intval(date('H'))-1;
		if ($lastFullHour > 0) {
			$sumTodayNbv	= 0;
			$sumTodayNba	= 0;
			$sumTodayBc		= 0;
			if (is_array($aToday)) {
				foreach ($aToday as $value) {
					if (is_object($value)) {
						if ($value->getColumn('label')>$lastFullHour) {
							continue;
						}
						$sumTodayNbv += $value->getColumn(Piwik_Archive::INDEX_NB_VISITS);
						$sumTodayNba += $value->getColumn(Piwik_Archive::INDEX_NB_ACTIONS);
						$sumTodayBc += $value->getColumn(Piwik_Archive::INDEX_BOUNCE_COUNT);
					}
				}
			}
			$sumLastweekNbv	= 0;
			$sumLastweekNba = 0;
			$sumLastweekBc	= 0;
			$sumAllLastweekNbv	= 0;
			$sumAllLastweekNba	= 0;
			$sumAllLastweekBc	= 0;
			if (is_array($aLastweek)) {
				foreach ($aLastweek as $value) {
					if (is_object($value)) {
						if ($value->getColumn('label')>$lastFullHour) {
							$sumAllLastweekNbv += $value->getColumn(Piwik_Archive::INDEX_NB_VISITS);
							$sumAllLastweekNba += $value->getColumn(Piwik_Archive::INDEX_NB_ACTIONS);
							$sumAllLastweekBc += $value->getColumn(Piwik_Archive::INDEX_BOUNCE_COUNT);
							continue;
						}
						$sumLastweekNbv += $value->getColumn(Piwik_Archive::INDEX_NB_VISITS);
						$sumLastweekNba += $value->getColumn(Piwik_Archive::INDEX_NB_ACTIONS);
						$sumLastweekBc += $value->getColumn(Piwik_Archive::INDEX_BOUNCE_COUNT);
					}
				}
			}
			$gradientTodayNbv		= ($sumTodayNbv-$sumLastweekNbv)/($this->today-$lastweek);
			$gradientTodayNba		= ($sumTodayNba-$sumLastweekNba)/($this->today-$lastweek);
			$gradientTodayBc		= ($sumTodayBc-$sumLastweekBc)/($this->today-$lastweek);

			$this->timefactor_nb_visits				= $gradientTodayNbv;
			$this->timefactor_nb_actions			= $gradientTodayNba;
			$this->timefactor_bounce_count			= $gradientTodayBc;
		} else {
			$this->timefactor_nb_visits				= 1;
			$this->timefactor_nb_actions			= 1;
			$this->timefactor_bounce_count			= 1;
		}
	}

	private function getForecastValues($idSite, $toFetch) {
		$num = 0;

		$today		= strtotime('today');
		$tomorrow	= strtotime('tomorrow');

		$numToday	= $this->getNumeric($idSite, 'day', date('Y-m-d', $today), $toFetch); // base Value
		$oneday		= 86400;
		$oneweek	= 7*$oneday;
		$twoweek	= 2*$oneweek;

		// Fetch the last 7 days of the same day
		$lastWeek	= array();
		$end		= strtotime('-7weeks', $today);
		for ($day=strtotime('-1week', $today); $day>=$end; $day=strtotime('-1week', $day)) {
			$lastWeek[$day] = $this->getNumeric($idSite, 'day', date('Y-m-d', $day), $toFetch);
		}

		// find the gradient
		$gradientLastWeeks = $this->findGradient($lastWeek);

		$timefactor = 0;
		if (property_exists($this, 'timefactor_'.$toFetch)) {
			eval('$timefactor = $this->timefactor_'.$toFetch.';');
		} else {
			$timefactor = $this->timefactor_nb_visits;
		}

		// *****************
		// calculate
		$numLastWeek		= $lastWeek[strtotime('-1week', $today)];
		// average growth of the last 7 days of the same weekday (overall values)
		$numChangeAverage	= $gradientLastWeeks*($tomorrow-time());
		// growth compared to the last week on same day (only to the last hour)
		$numChangeLastWeek	= $timefactor*$oneweek;

		// forecasted value
		$forecast	= $numLastWeek + $numChangeAverage + $numChangeLastWeek;
		$num		= ($forecast>$numToday?$forecast:$numToday+1);

		// percentage of change to the last week on same day
		$gradient	= ceil(($lastWeek[strtotime('-1week', $today)]==0?100:$num*100/$lastWeek[strtotime('-1week', $today)]));

		$icon = 'FA';
		switch ($toFetch) {
			case 'bounce_count':
				if ($gradient>=100) {
					$icon .= 'Red';
				} else {
					$icon .= 'Green';
				}
				break;
			default:
				if ($gradient>=100) {
					$icon .= 'Green';
				} else {
					$icon .= 'Red';
				}
				break;
		}

		if ($gradient>115) {
			$icon .= 'StraightUp';
		} else if ($gradient>101) {
			$icon .= 'Up';
		} else if ($gradient<85) {
			$icon .= 'StraightDown';
		} else if ($gradient<99) {
			$icon .= 'Down';
		} else {
			$icon .= 'Side';
		}

		$path	= dirname($_SERVER['REQUEST_URI']);
		return '<div title="'.($gradient>100?'+':($gradient==100?'':'-')).' '.($gradient>100?$gradient-100:100-$gradient).'%"><div class="forecastArrows '.$icon.'">&nbsp;</div>'.ceil($num).'<div style="clear:left" /></div>';
	}

	private function findGradient($arr) {
		$lastValue = null;
		$lastKey = null;
		$temp = 0;
		foreach ($arr as $key => $value) {
			if (!is_null($lastValue)) {
				$temp += ($lastValue-$value)/($lastKey-$key);
			}
			$lastValue = $value;
			$lastKey = $key;
		}
		$gradient = $temp/(count($arr)-1);

		return $gradient;
	}

	protected function getNumeric( $idSite, $period, $date, $toFetch )
	{
		Piwik::checkUserHasViewAccess( $idSite );
		$archive = Piwik_Archive::build($idSite, $period, $date );
		$dataTable = $archive->getNumeric($toFetch);
		return $dataTable;
	}

	protected function getBlob( $idSite, $period, $date, $toFetch )
	{
		Piwik::checkUserHasViewAccess( $idSite );
		$archive = Piwik_Archive::build($idSite, $period, $date );
		$dataTable = $archive->getBlob($toFetch);
		return unserialize($dataTable);
	}
}
